#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['falcon_caching', 'falcon_caching.backends']

package_data = \
{'': ['*']}

install_requires = \
['falcon >=1.0.0']

extras_require = \
{'test': ['flit',
          'pylibmc',
          'pytest >=4.0.0',
          'pytest-cov',
          'pytest-pep8',
          'pytest-xdist',
          'pytest-xprocess',
          'redis',
          'tox',
          'tox-pipenv-install']}

setup(name='falcon_caching',
      version='0.1.0',
      description='Falcon-Caching - a caching module for the Falcon web framework',
      author='Zoltan Fedor',
      author_email='zoltan.1.fedor@gmail.com',
      url='https://github.com/zoltan-fedor/falcon-caching',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
